package com.lemoulinstudio.jnag.utils.rds.client;

import com.sun.sgs.client.ClientChannel;
import com.sun.sgs.client.ClientChannelListener;
import com.sun.sgs.client.simple.SimpleClient;
import com.sun.sgs.client.simple.SimpleClientListener;
import java.io.IOException;
import java.net.PasswordAuthentication;
import java.nio.ByteBuffer;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import com.lemoulinstudio.jnag.jse.AbstractConfiguration;
import com.lemoulinstudio.jnag.jse.JnagSession;
import com.lemoulinstudio.jnag.jse.JnagSessionImpl;

/**
 * <p>A class that can be helpful to setup a client app that is using Jnag and
 * the Red Dwarf Server.</p>
 *
 * <p>Jnag samples are all using it for convenience.
 * For your app, feel free to use it or not.</p>
 *
 * @param <AppContext> The class used by the client app to store objects
 * global to the application.
 *
 * The goal of this <code>AppContext</code> class is to not have static
 * fields in the program, so that we can run many instances of 1 client
 * inside of the same Java Virtual Machine.
 *
 * @author Vincent Cantin
 */
public abstract class GenericClient<AppContext> {

  private class ConnectionListener implements SimpleClientListener {
    
    public void login(SimpleClient client) {
      Properties connectProps = new Properties();
      connectProps.put("host", serverHostName);
      connectProps.put("port", serverPortName);
      
      try {
        client.login(connectProps);
      } catch (IOException ex) {
        Logger.getLogger(GenericClient.class.getName()).log(Level.SEVERE, null, ex);
      }
    }

    @Override
    public PasswordAuthentication getPasswordAuthentication() {
      return new PasswordAuthentication(userName, "pass".toCharArray());
    }

    @Override
    public void loggedIn() {
      notifyConnectedToServer();
    }

    @Override
    public void loginFailed(String reason) {
      Logger.getLogger(GenericClient.class.getName()).log(Level.SEVERE, null, reason);
    }

    @Override
    public ClientChannelListener joinedChannel(ClientChannel clientChannel) {
      return null;
    }

    @Override
    public void receivedMessage(ByteBuffer message) {
      //System.out.println("Recv " + Utils.byteBufferToString(message));
      jnagSession.decodeAndExecute(message);
    }

    @Override
    public void reconnecting() {
    }

    @Override
    public void reconnected() {
      // Not yet handled / experimented / needed.
      // notifyConnectedToServer();
    }

    @Override
    public void disconnected(boolean b, String reason) {
      notifyDisconnectedFromServer();
    }
    
  }

  
  private String serverHostName;
  private String serverPortName;
  private String userName = "user" + ((int) (Math.random() * 10000));
  private JnagSession jnagSession;
  private AppContext appContext;

  /**
   * Creates an instance of a GenericClient which will connect the server
   * using the default host name "localhost" and the default port number "1139".
   *
   * @param config An instance of the config class generated by the annotation processor.
   */
  public GenericClient(AbstractConfiguration config) {
    this("localhost", "1139", config);
  }

  /**
   * Creates an instance of a GenericClient which will connect the server.
   *
   * @param serverHostName The name of the host where the server is localted.
   * 
   * @param serverPortName The port of the server app.
   *
   * @param config An instance of the config class generated by the annotation processor.
   */
  public GenericClient(String serverHostName, String serverPortName, AbstractConfiguration config) {
    this.serverHostName = serverHostName;
    this.serverPortName = serverPortName;

    ConnectionListener connectionListener = new ConnectionListener();
    SimpleClient simpleClient = new SimpleClient(connectionListener);
    
    jnagSession = new JnagSessionImpl(config);
    jnagSession.setMessageListener(new MessageSenderToServerSession(simpleClient));

    // Setup the app context. The JnagSession is available.
    appContext = createAppContext();

    // Setup some objects in the jnag session for communication bootstrap.
    setupCommunicationBootstrap();

    // Connect the server.
    connectionListener.login(simpleClient);
  }

  /**
   * Returns the app context.
   *
   * @return The app context.
   */
  public AppContext getAppContext() {
    return appContext;
  }

  /**
   * Returns the user name that have been randomly generated for
   * this generic client. It is used when login into the server.
   *
   * @return The user name.
   */
  public String getUserName() {
    return userName;
  }

  /**
   * Returns the jnag session created by this generic client.
   * 
   * @return The jnag session.
   */
  protected JnagSession getJnagSession() {
    return jnagSession;
  }

  /**
   * This creates the instance of the app context.
   *
   * The default implementation returns null.
   * Override this function if you want to specify your own instance.
   *
   * @return a new app context instance.
   */
  protected AppContext createAppContext() {return null;}

  /**
   * Called after the Jnag session have been created and before
   * the connection attempt to the server.
   *
   * That's the right place to setup the Jnag session, its outputs, proxies,
   * expose objects, etc ...
   */
  protected abstract void setupCommunicationBootstrap();

  /**
   * Called when we are connected to the server.
   *
   * The default implementation does nothing.
   * You can start to send messages after you receive this event.
   */
  protected void notifyConnectedToServer() {}

  /**
   * Called when we are disconnected from the server.
   *
   * The default implementation does nothing.
   * You should stop to send messages after you receive this event.
   */
  protected void notifyDisconnectedFromServer() {}

}
